-- Examples used in the Fran tutorial, which may be found at:
-- 
--   http://www.research.microsoft.com/~conal/fran/tutorial.hs
-- 
-- Instructions (also found in tutorial.hs):
-- 
--   After installing Hugs from [Hugs 1.4], double-click on the file
--   tutorial.hs in the subdirectory lib\Fran\demos.  Run the examples by
--   typing "main ()". Press space, "n", or right-arrow to advance to
--   the next animation, and "p" or left-arrow to back up to the previous
--   one. If you want to display just a single animation, e.g.,
--   leftRightCharlotte, then close the animation window and type "display
--   leftRightCharlotte" to the prompt. You can alter the definition in an
--   editor, save the result, type ":r" to the Hugs prompt, and "$$" again
--   to display the new version. For examples involving user interaction or
--   rate-based animation, use displayT instead of display, as mentioned
--   below.


module Tutorial where

import Fran     -- Bring RBMH into scope

main _ = displayT allAnims


importKid name = importBitmap ("../Media/" ++ name ++ " head.bmp")


leftRightCharlotte = moveXY wiggle 0 charlotte

charlotte = importKid "charlotte"


upDownPat = moveXY 0 waggle pat

pat = importKid "pat" 


charlottePatDance = leftRightCharlotte `over` upDownPat 


hvDance im1 im2 =
  moveXY wiggle 0 im1  `over`
  moveXY 0 waggle im2


charlottePatDance2 = hvDance charlotte pat


charlottePatDoubleDance = hvDance aSmall aSmall
  where
   aSmall = smaller 2 charlottePatDance


dance1 = bigger (abs wiggle) charlottePatDance


dance2 = hvDance (bigger wiggle charlotte)
                 (bigger waggle pat)


patOrbitsCharlotte = bigger wiggle charlotte       `over`
                     moveXY wiggle waggle pat

-- Rate-based animation

velBecky t0 = moveXY x 0 becky
  where
    x = -1 + atRate 1 t0

becky = importKid "becky"

accelBecky t0 = moveXY x 0 becky
  where
    x = -1 + atRate v t0
    v =  0 + atRate 1 t0

mouseVelBecky t0 = move offset becky
  where
   offset = atRate (mouseMotion t0) t0 

beckyChaseMouse t0 = move offset becky
  where
   offset = atRate vel t0
   vel    = mouseMotion t0 - offset

chaseMouse im t0 = move offset im
  where
   offset = atRate vel t0
   vel    = mouseMotion t0 - offset

danceChase t0 =
  chaseMouse (smaller 2 charlottePatDance) t0

springDragBecky t0 = move offset becky
  where
   offset = atRate vel t0
   vel    = atRate accel t0
   accel  = (mouseMotion t0 - offset) - 0.5 *^ vel

-- Composition in Time

orbitAndLater = orbit `over` later 1 orbit
  where
   orbit = moveXY wiggle waggle jake

jake = importKid "jake" 

orbitAndSlower = orbit `over` slower 2 orbit
  where
   orbit = moveXY wiggle waggle jake

-- For orbitAndSlower and followMouseAndDelay, jake looks silly because
-- there's no transparency in this version.

followMouseAndDelay t0 = follow `over` later 1 follow
  where
   follow = move (mouseMotion t0) jake

delayAnims dt anims =
  overs (zipWith (later . constantB) [0, dt ..] anims)

-- This one will be better when background transparency works

kids t0 =
  delayAnims 0.5 (map (move (mouseMotion t0))
                      [jake, becky, charlotte, pat])

trailWords motion str =
  delayAnims 1 (map moveWord (words str))
  where
    moveWord word = move motion (
                     bigger 2 (
                       withColor blue (stringIm word) ))

flows t0 = trailWords motion "Time flows like a river"
  where
    motion = 0.7 *^ vector2XY (cos time) (sin (2 * time))

flows2 t0 = trailWords (mouseMotion t0) "Time flows like a river"



-- Reactive Animation

redBlue t0 = buttonMonitor t0 `over`
             withColor c circle
  where
   c = red `untilB` lbp t0 -=> blue


redBlueCycle t0 = buttonMonitor t0 `over`
                  withColor (cycle red blue t0) circle
  where
   cycle c1 c2 t0 =
    c1 `untilB` lbp t0 *=> cycle c2 c1


tricycle t0 =
   buttonMonitor t0 `over`
   withColor (cycle3 green yellow red t0) (
     bigger (wiggleRange 0.5 1)
       circle )
  where
   cycle3 c1 c2 c3 t0 =
    c1 `untilB` lbp t0 *=> cycle3 c2 c3 c1


jumpFlower t0 = buttonMonitor t0 `over`
                moveXY (bSign t0) 0 flower

flower = smaller 2.5 (importBitmap "../Media/rose medium.bmp")

selectLeftRight :: a -> a -> a -> Time -> Behavior a

selectLeftRight none left right t0 = notPressed t0
 where
  notPressed t0 =
   constantB none `untilB` 
     lbp t0 *=> pressed left  lbr .|. 
     rbp t0 *=> pressed right rbr 

  pressed x stop t0 = 
   constantB x `untilB` stop t0 *=> notPressed

bSign :: Time -> RealB

bSign = selectLeftRight 0 (-1) 1

buttonMonitor t0 =
  moveXY 0 (-1) (
   withColor black (
    bigger 2 (
    stringBIm (selectLeftRight "(press a button)" "left" "right" t0))))
 

growFlower t0 = buttonMonitor t0 `over`
                grow flower t0

grow im t0 = bigger size im 
 where 
  size = 1 + atRate rate t0 
  rate = bSign t0 


growFlowerExp t0 = buttonMonitor t0 `over`
                   grow' flower t0

grow' im t0 = bigger size im 
 where 
  size = 1 + atRate rate t0 
  rate = bSign t0 * size 


-- A fun alternative to "flower" in the previous examples:
godzilla = moveXY 0 (-0.4) (
           smaller 2.5 (importBitmap "../Media/godzilla medium.bmp") )



-- --- Misc to go elsewhere ---


-- Mouse motion vector, i.e., where the mouse is relative to the origin

mouseMotion :: Time -> Vector2B

mouseMotion t0 = mouse t0 .-. origin2

delay = later

overs :: [ImageB] -> ImageB

overs = foldr over emptyImage

display imB = disp (\ t0 -> delay (constantB t0) imB)

displayT = disp

displaySave dir dt imB = dispSave dir dt (const imB)


-- Packing it all up

titles = [ "leftRightCharlotte", "upDownPat", "charlottePatDance",
          "charlottePatDoubleDance", "dance1", "dance2",
          "patOrbitsCharlotte",
          "velBecky", "accelBecky", "mouseVelBecky", "beckyChaseMouse",
          "danceChase",
          "springDragBecky", "orbitAndLater", "orbitAndSlower",
          "followMouseAndDelay", "flows", "flows2",
          "redBlue", "redBlueCycle", "tricycle",
          "jumpFlower", "growFlower", "growFlowerExp",
          "The End" ]

anims = map const [ talkTitle,
                    leftRightCharlotte, upDownPat, charlottePatDance,
                    charlottePatDoubleDance, dance1, dance2,
                    patOrbitsCharlotte ]
        ++
        [ velBecky, accelBecky, mouseVelBecky, beckyChaseMouse, danceChase,
          springDragBecky, const orbitAndLater, const orbitAndSlower,
          followMouseAndDelay, flows, flows2,
          redBlue, redBlueCycle, tricycle,
          jumpFlower, growFlower, growFlowerExp,
          const emptyImage ]

{-
animsAndTitles =
 concat [ [titleIm title, anim]
         | (title,anim) <- zip titles anims]
  where
   titles = [ "leftRightCharlotte", "upDownPat", "charlottePatDance",
              "charlottePatDoubleDance", "dance1", "dance2",
              "patOrbitsCharlotte",
              "velBecky", "accelBecky", "mouseVelBecky", "beckyChaseMouse",
              "danceChase",
              "springDragBecky", "orbitAndLater", "orbitAndSlower",
              "followMouseAndDelay", "flows", "flows2",
              "redBlue", "redBlueCycle", "tricycle",
              "jumpFlower", "growFlower", "growFlowerExp",
              "The End" ]

   titleIm name t0 =  bigger (25 / fromInt (length name)) (
                        withColor red (stringIm name) )
-}

animsAndTitles = interleave anims animTitles
 where

  animTitles =
   [  later' (tAnim (tsize str (withColor col (stringIm str))))
    | (col,str,tAnim) <- zip3 colors titles
                              (concat (repeat titleAnimators)) ]

  colors = [ hsl (constantB h + 30*wiggle) 0.5 0.5 | h <- [0, sep .. 359] ]

  sep = fromInt 360 / fromInt (length titles)

  tsize str = bigger (25 / fromInt (length str))

  interleave (x:xs') ys = x : interleave ys xs'
  interleave _       _  = []

  later' i t = later (constantB t) i

allAnims t0 = allRec t0 0
 where
  allRec t0 i = (animsAndTitles!!i) t0 `untilB` newIndex +=> allRec
    where
      newIndex = (plusMinus `suchThat` validIndex) t0
      validIndex i'  =  0 <= i' && i' < lenAnims
      plusMinus t0 = keyIn " N'" t0 -=> i+1  .|. -- "<space>n<right>"
                     keyIn "P%"  t0 -=> i-1      -- "<left>p"
      keyIn chars =
        keyPress `suchThat` (\(ch,_) -> ch `elem` chars)
  lenAnims = length animsAndTitles


talkTitle =
 bigger 1.5 $
 trail 1.2 motion
   [ withColor (hsl (constantB h + wiggleColor) 0.5 0.5) (
       bigger 1.5 (stringIm str))
   | (str,h) <- zip titleWords
                    [0, 360 / fromIntegral (length titleWords) .. 359]  ]
 where
  motion = point2XY (slower 5 (0.3 * waggle))
                    (slower 4 (0.3 * wiggle))
  wiggleColor = slower 2 (180 * wiggle)
  titleWords = words "Functional Reactive Animation"

  trail dt motion anims =
    delayAnims dt (map (move motion) anims)

-- Now let's look at some simple title animators

rockT,growT,orbitT :: ImageB -> ImageB

-- First, a rocking motion, in which we move down, rotate back and forth
-- slightly, and move back up.

rockT = moveXY 0 0.1             .
        turnRight (wiggle / 10)  .
        moveXY 0 (-0.2)

-- Then a pulsating scaling:

growT = bigger (wiggleRange 0.7 1.3)

-- And finally a small circular motion:

orbitT = moveXY (0.3*wiggle) (0.3*waggle)

-- To compose the title animators we double up the simple ones plus the
-- identity animator.

titleAnimators = [ a . a' | a  <- simples, a'  <- simples ]
 where
  simples = [id, rockT, growT, orbitT]
